/* Derived from test_apogee.cpp                                             */
/*                                                                          */
/*  To build for ALTA-U                                                     */
/*                                                                          */
/*  g++ -I/opt/apogee/include  -c apogeedrv.cpp -IFpgaRegs -o apogeedrv \   */
/*    ApnCamera.o ApnCamera_Linux.o ApnCamera_USB.o  \                      */
/*    ApogeeUsbLinux.o ApnCamData*.o ApnCamTable.o  \                       */
/*    -lusb                                                                 */
/*                                                                          */

#include <sys/types.h>
#include <sys/shm.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "fitsio.h"
#include <math.h>
#include <unistd.h>
#include <time.h>

#include "ApnCamera.h"
#include "apogeedrv.h"

CApnCamera *alta;

typedef struct {
     unsigned short *pixels;
     int            size;
     short          xdim;
     short          ydim;
     short          zdim;
     short          xbin;
     short          ybin;
     short          type;
     char           name[64];
     int            shmid;
     size_t         shmsize;
     char           *shmem;
} CCD_FRAME;
#define MAX_CCD_BUFFERS  1000
CCD_FRAME CCD_Frame[MAX_CCD_BUFFERS];

typedef void *PDATA;
PDATA CCD_locate_buffer(char *name, int idepth, short imgcols, short imgrows, 
  short hbin, short vbin);

int CCD_locate_buffernum(char *name);
int CCD_free_buffer(char *name);


/* CCD utilities from ccd_tcl.c */

int CCD_locate_buffernum(char *name)
{
  int found;
  int i;

  found = -1;
  i = 0;
 
  while (found<0 && i<MAX_CCD_BUFFERS) 
  {

    if (CCD_Frame[i].pixels != NULL) 
    {
      if (strcmp(name,CCD_Frame[i].name) == 0) 
      {
         found = i;
      } 
    }
    i++;
  }
  
  return(found);
}

int CCD_free_buffer(char *name)
{
  int found;
  int i;

  found = -1;
  i = 0;
 
  while (found<0 && i<MAX_CCD_BUFFERS) 
  {

   if (CCD_Frame[i].pixels != NULL) 
   {
     if (strcmp(name,CCD_Frame[i].name) == 0) 
     {
       found = i;
     } 
   }
   i++;
  }
  
  if (found >= 0) {
    free(CCD_Frame[found].pixels);
    CCD_Frame[found].pixels = NULL;
    CCD_Frame[found].size = 0;
    strcpy(CCD_Frame[found].name,"NOT-IN-USE");
    if (CCD_Frame[found].shmem != NULL) 
    {
      shmdt(CCD_Frame[found].shmem);
      CCD_Frame[found].shmem = NULL;
    }
  }
  return(0);
}

extern "C" {

static int exp_x = 0;
static int exp_y = 0;
static int exp_width = 0;
static int exp_height = 0;

int apogee_init_camera (void) 
{
  int status;
  int camnum;

  /* Create the camera object , this will reserve memory */
  
  alta = (CApnCamera *)new CApnCamera();
  camnum = 1;
  status = alta->InitDriver(camnum,0,0);
  alta->ResetSystem();
  if (!status)
  {    
    /* Close and try again */
    alta->CloseDriver();
    usleep(2000000);
    status = alta->InitDriver(camnum,0,0);
    if (!status)
    {
      return(0);
    }
  }
  
  /* No binning */

  alta->write_RoiBinningH(1);
  alta->write_RoiBinningV(1);
  
  /* 16-bit acquisition */
  
  alta->write_DataBits (Apn_Resolution_SixteenBit);
  
  /* Fan on*/
  
  alta->write_FanMode(Apn_FanMode_High); 
      
  /* Returns 1 on success and 0 on failure to initialize camera */
  
  return(1);
}


int apogee_close_camera(void)
{
  if (alta)
  {
    alta->CloseDriver();
    delete alta;
  }
  return(1);
}

int apogee_camera_present(int *type, int *width, int *height)
{ 
  int status;
  int ccd_type, ccd_image_width, ccd_image_height;
  
  #ifndef APOGEE
  #define APOGEE 0xFFE0  
  #endif
  #ifndef NO_CAMERA
  #define NO_CAMERA 0xFFFF 
  #endif
  
  ccd_type = NO_CAMERA;
  ccd_image_height = 0;
  ccd_image_width = 0;
  
  status = alta->read_Present();
  fprintf(stderr,"read_Present()=%d\n", status);
   
  if (!status)
  {
    *type = ccd_type;
    *height = ccd_image_height;
    *width = ccd_image_width;
    return(0);
  }

  ccd_type = APOGEE;
  ccd_image_height = alta->read_RoiPixelsH();
  ccd_image_width = alta->read_RoiPixelsV();
      
  *type = ccd_type;
  *height = ccd_image_height;
  *width = ccd_image_width;
 
  return(1);
}


int apogee_camera_status(void)
{
  Apn_Status imaging;

  #ifndef IDLE 
  #define IDLE 0
  #endif
  #ifndef INTEGRATING
  #define INTEGRATING 1
  #endif
  #ifndef COMPLETE
  #define COMPLETE 2
  #endif  
  
  imaging = alta->read_ImagingStatus();
  if (imaging  < 0)
  {
    return(IDLE);
  }
  else if (imaging ==  Apn_Status_ImageReady)
  {
    return(COMPLETE);
  }
  
  return(IDLE);
}     
  


int apogee_get_image(unsigned short int *image)
{  

  long status;
  unsigned long count;
  unsigned short width;
  unsigned short height;
  
  width = exp_width;
  height = exp_height;
  status = alta->GetImageData (image, width, height, count);
  if (status != CAPNCAMERA_SUCCESS)
  {
    return(0);
  }       

  return(1);
}


int apogee_expose(double exposure, int x, int y, int height, int width, 
  int shutter_flag)
{
  int status;
  
  exp_width = width;
  exp_height = height;
  exp_x = x;
  exp_y = y;
  
  alta->write_RoiStartX(x);
  alta->write_RoiStartY(y);
  alta->write_RoiPixelsH(width);
  alta->write_RoiPixelsV(height);
  
  /* Single image per download */
  alta->write_ImageCount(1);

  /* Start an exposure */
  status = alta->Expose(exposure,shutter_flag);
  
  if (!status)
  {
    return(0);
  }
    
  return(1);
}
  

int apogee_stop_exposure(int readout_flag)
{
  if (readout_flag == FALSE)
  {
    /* No readout expected */
    alta->StopExposure(FALSE);
    return(1);
  }
  else
  {
    /* Subsequent readout expected */
    alta->StopExposure(TRUE);
    return(1);
  }    
}  
      

int apogee_guide(int relay, double ontime)
{
  /* Controlling guider is undocumented and not yet supported in XmCCD */
  return(0);
}

int apogee_tiptilt(int ao_x, int ao_y)
{
  /* There is no tip-tilt control yet available for this camera */
  return(0);
}


int apogee_cooler(int set_flag, double setpoint)
{
  alta->write_CoolerEnable(set_flag);
  if (set_flag)
  {
    alta->write_CoolerEnable(TRUE);
    alta->write_FanMode(Apn_FanMode_High);
    if (setpoint > -60.)
    {
      alta->write_CoolerSetPoint(setpoint);
    }
    else
    {
       alta->write_CoolerSetPoint(-60.); 
    }   
  }
  else
  {
    alta->write_CoolerEnable(FALSE);
    alta->write_FanMode(Apn_FanMode_Low);
    alta->write_CoolerSetPoint(100);
  }    
  return(1); 
}


int apogee_temperature(double *power, double *temperature)
{
  double air_temperature;
  *power = alta->read_CoolerStatus();
  *temperature = alta->read_TempCCD();
  air_temperature = alta->read_TempHeatsink ();
  return(1);
}

}
